﻿using System;
using System.Text.RegularExpressions;

namespace nick3216.Library.Crypt.Password
{
    /// <summary>
    /// <para>Generate  random passwords that are (more or less) pronounceable. It 
    /// doesn't use the FIPS-181 NIST standard, which is proven to be 
    /// insecure.</para>
    /// 
    /// <para>nick3216.Library.Crypt.Password.Generate supports English and German.</para>
    /// 
    /// <para>This library improves on FIPS-181 using a true random selection with 
    /// the word generator as mere filter. Other improvements are better 
    /// pronounceability using third order approximation instead of second 
    /// order and multi-language support. The drawback of this method is that 
    /// it is usually slower.</para>
    /// 
    /// <para>For details on why FIPS-181 is insecure and why the solution used in this 
    /// module is reasonably secure, see "A New Attack on Random Pronounceable 
    /// Password Generators" by Ravi Ganesan and Chris Davies, available online 
    /// in may places - use your favourite search engine.</para>
    /// 
    /// <h1>Authors</h1>
    /// <para>Nick Wallis &lt;<a href='mailto:singlespeeder@btinternet.com'>singlespeeder@btinternet.com</a>&gt;<br/>
    /// Based on the perl <a href='http://search.cpan.org/~jwalt/Crypt-GeneratePassword-0.03/'>Crypt::GeneratePassword</a>
    /// module by Jörg Walter.<br/>
    /// Inspired by ideas from Tom Van Vleck and Morris Gasser/FIPS-181.</para>
    /// 
    /// <h1>Copyright</h1>
    /// <para>Copyright (C) 2009, Nick Wallis &lt;<a href='mailto:singlespeeder@btinternet.com'>singlespeeder@btinternet.com</a>&gt;<br/>
    /// Portions Copyright (C) 2002, Jörg Walter &lt;<a href='mailto:jwalt@cpan.org'>jwalt@cpan.org</a>&gt;<br/>
    /// All rights reserved.</para>
    ///
    /// <para>This file is part of nick3216.Library.Crypt.Password.
    /// nick3216.Library.Crypt.Password is free software: you can 
    /// redistribute it and/or modify it under the terms of the GNU General 
    /// Public License as published by the Free Software Foundation, either 
    /// version 3 of the License, or (at your option) any later version.</para>
    ///
    /// <para>nick3216.Library.Crypt.Password is distributed in the hope that it 
    /// will be useful, but WITHOUT ANY WARRANTY; without even the implied 
    /// warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    /// See the GNU General Public License for more details.</para>
    ///
    /// <para>You should have received a copy of the GNU General Public License
    /// along with nick3216.Library.Crypt.Password.  
    /// If not, see <a href='http://www.gnu.org/licenses/'>http://www.gnu.org/licenses/</a></para>
    /// </summary>
    public class Generator
    {
        /// <summary>
        /// The default language for the generator.
        /// </summary>
        private ILanguage m_language = new en();

        // Use a globally visible randomiser which is seeded on object 
        // construction.
        /// <summary>
        /// The random number generator.
        /// </summary>
        private Random m_randomNumber = new Random();

        /// <summary>
        /// Seeds this instance.
        /// If the generator is seeded with milliseconds then different 
        /// objects created at the same time will return the same result, 
        /// and the results will repeat on average every 1000 iterations.
        /// Instead use the Crypto Service Random Number Generator to 
        /// generate the seed
        /// </summary>
        private void Seed()
        {
            byte[] randomBytes = new byte[4];
            System.Security.Cryptography.RNGCryptoServiceProvider rng =
                new System.Security.Cryptography.RNGCryptoServiceProvider();
            rng.GetBytes(randomBytes);
            int seed = (randomBytes[0] & 0x7f << 24 |
                randomBytes[1] << 16 |
                randomBytes[2] << 8 |
                randomBytes[3]);
            m_randomNumber = new Random();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Generator"/> class.
        /// </summary>
        public Generator()
        {
            Seed();
        }

        /// <summary>
        /// Initializes a new instance of the <see cref="Generator"/> class 
        /// for the specified language.
        /// </summary>
        public Generator(ILanguage language)
        {
            Seed();
            m_language = language;
        }

        /// <summary>
        /// Generates a completely random word between minLength and maxLength
        /// long.
        /// Length of the result is determined by relative probability not 
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Minimum length of the result.</param>
        /// <param name="maxLength">maximum length of the result.</param>
        /// <returns>A string of random characters.</returns>
        public String GetCharacters(int minLength, int maxLength)
        {
            return this.GetCharacters(
                minLength, maxLength, m_language.DefaultCharacterSet);
        }

        /// <summary>
        /// Generates a completely random word between minLength and maxLength
        /// long, using the specified character set.
        /// Length of the result is determined by relative probability not
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Minimum length of the result.</param>
        /// <param name="maxLength">maximum length of the result.</param>
        /// <param name="set">String containing the set of characters to use.</param>
        /// <returns>A string of random characters from the specified set.</returns>
        public String GetCharacters(int minLength, int maxLength, String set)
        {
            return this.GetCharacters(
                minLength, maxLength, set, m_language.Filter);
        }

        /// <summary>
        /// Generates a completely random word between minLength and maxLength
        /// long, using the specified character set, and with the results 
        /// filtered for potential offensive sequences.
        /// Length of the result is determined by relative probability not
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Length of the min.</param>
        /// <param name="maxLength">Length of the max.</param>
        /// <param name="set">String containing the set of characters to use.</param>
        /// <param name="filter">String containing a regular expression filter
        /// of character sequences to filter out of results, e.g. potentially
        /// offensive sequences.</param>
        /// <returns>A filtered string of random characters from the specified 
        /// set.</returns>
        public String GetCharacters(
            int minLength, int maxLength, String set, String filter)
        {
            // Add the null character into the supplied set. This will be 
            // removed from any generated passwords, which is how the length 
            // is randomised.
            set = set + '\x00';

            if (minLength < 1)
            {
                minLength = 1;
            }
            if (maxLength < minLength)
            {
                maxLength = minLength;
            }

            String rawPassword;
            String password;
            do
            {
                // Must reset results each time the loop is entered or each 
                // loop just adds to the end of the result, which is no good 
                // if the loop is rerunning because of an offensive character 
                // sequence which will just be rematched.
                rawPassword = "";

                // Construct a string of maxlength random characters from the allowed set
                for (int i = 0; i < maxLength; i++)
                {
                    int j = m_randomNumber.Next(0, set.Length);
                    rawPassword = rawPassword + set[j];
                }

                // Replace null characters in the password between with nothing
                // at all, which will shorten the result (hence the differences 
                // in length).
                Regex nullChar = new Regex('\x00'.ToString());
                password = nullChar.Replace(rawPassword, "");

            } while (IsRude(password, filter) || password.Length < minLength);

            return password;
        }

        /// <summary>
        /// Generates a random, pronounceable word between minLength and 
        /// maxLength long.
        /// Pronounceability is based on occurence of four character sequences 
        /// in the password generator language.
        /// Length of the result is determined by relative probability not
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Minimum length of the password.</param>
        /// <param name="maxLength">Maximum length of the password.</param>
        /// <returns>A random pronounceable word.</returns>
        public String GetWord(int minLength, int maxLength)
        {
            return GetWord(minLength, maxLength, false, false);
        }

        /// <summary>
        /// Generates a random, pronounceable word between minLength and 
        /// maxLength long.
        /// Pronounceability is based on occurence of four character sequences 
        /// in the password generator language.
        /// Length of the result is determined by relative probability not
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Minimum length of the password.</param>
        /// <param name="maxLength">Maximum length of the password.</param>
        /// <param name="useSymbols">If set to <c>true</c> then allow symbols 
        /// in the result.</param>
        /// <param name="useCapitals">If set to <c>true</c> then allow capital
        /// letters in the result.</param>
        /// <returns>A random pronounceable word.</returns>
        public String GetWord(
            int minLength, int maxLength, bool useSymbols, bool useCapitals)
        {
            // Uses default frequencies for quad sequences
            return GetWord(minLength, maxLength, useSymbols, useCapitals, 0.0001, 0.001);
        }

        /// <summary>
        /// Generates a random, pronounceable word between minLength and 
        /// maxLength long.
        /// Pronounceability is based on occurence of four character sequences 
        /// in the password generator language.
        /// Length of the result is determined by relative probability not
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Minimum length of the password.</param>
        /// <param name="maxLength">Maximum length of the password.</param>
        /// <param name="useSymbols">If set to <c>true</c> then allow symbols 
        /// in the result.</param>
        /// <param name="useCapitals">If set to <c>true</c> then allow capital 
        /// letters in the result.</param>
        /// <param name="minFrequency">The minimum allowed frequency of the 
        /// character sequences (from 0.0 to 1.0). A higher value uses more 
        /// common character sequences making the password more memorable, 
        /// at the expense of security.</param>
        /// <param name="averageFrequency">The average frequency of character 
        /// sequences(from 0.0 to 1.0). A higher value uses more common 
        /// character sequences making the password more memorable, at the 
        /// expense of security.</param>
        /// <returns>A random pronounceable word.</returns>
        public String GetWord(
            int minLength, int maxLength,
            bool useSymbols, bool useCapitals,
            double minFrequency, double averageFrequency)
        {
            return GeneratePassword(
                minLength, maxLength,
                useSymbols, useCapitals,
                minFrequency, averageFrequency,
                false);
        }

        /// <summary>
        /// Generates a random, pronounceable word between minLength and 
        /// maxLength long.
        /// Pronounceability is based on occurence of three character 
        /// sequences in the password generator language.
        /// Length of the result is determined by relative probability not
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Minimum length of the password.</param>
        /// <param name="maxLength">Maximum length of the password.</param>
        /// <returns>A random pronounceable word.</returns>
        public String GetWord3(int minLength, int maxLength)
        {
            return GetWord3(minLength, maxLength, false, false);
        }

        /// <summary>
        /// Generates a random, pronounceable word between minLength and 
        /// maxLength long.
        /// Pronounceability is based on occurence of three character 
        /// sequences in the password generator language.
        /// Length of the result is determined by relative probability not
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Minimum length of the password.</param>
        /// <param name="maxLength">Maximum length of the password.</param>
        /// <param name="useSymbols">If set to <c>true</c> then allow symbols in the result.</param>
        /// <param name="useCapitals">If set to <c>true</c> then allow capital letters in the result</param>
        /// <returns>A random pronounceable word.</returns>
        public String GetWord3(
            int minLength, int maxLength, bool useSymbols, bool useCapitals)
        {
            return GetWord(minLength, maxLength, useSymbols, useCapitals, 0.0001, 0.05);
        }

        /// <summary>
        /// Generates a random, pronounceable word between minLength and 
        /// maxLength long.
        /// Pronounceability is based on occurence of three character 
        /// sequences in the password generator language.
        /// Length of the result is determined by relative probability not
        /// uniformly.
        /// </summary>
        /// <param name="minLength">Minimum length of the password.</param>
        /// <param name="maxLength">Maximum length of the password.</param>
        /// <param name="useSymbols">If set to <c>true</c> then allow symbols in the result.</param>
        /// <param name="useCapitals">If set to <c>true</c> then allow capital letters in the result</param>
        /// <param name="minFrequency">The minimum allowed frequency of the 
        /// character sequences (from 0.0 to 1.0). A higher value uses more 
        /// common character sequences making the password more memorable, 
        /// at the expense of security.</param>
        /// <param name="averageFrequency">The average frequency of character 
        /// sequences(from 0.0 to 1.0). A higher value uses more common 
        /// character sequences making the password more memorable, at the 
        /// expense of security.</param>
        /// <returns>A random pronounceable word.</returns>
        public String GetWord3(
            int minLength, int maxLength,
            bool useSymbols, bool useCapitals,
            double minFrequency, double averageFrequency)
        {
            return GeneratePassword(
                minLength, maxLength,
                useSymbols, useCapitals,
                minFrequency, averageFrequency,
                true);
        }

        /// <summary>
        /// Constrains the frequency supplied by the user to a value between 0.0 and 1.0.
        /// </summary>
        /// <param name="frequency">The frequency to constrain.</param>
        /// <returns>A value between 0.0 and 1.0.</returns>
        private double ConstrainFrequency(double frequency)
        {
            if (frequency < 0.0)
            {
                frequency = 0.0;
            }
            else if (frequency > 1.0)
            {
                frequency = frequency - (int)frequency;
            }
            return frequency;
        }

        /// <summary>
        /// Generates a password.
        /// </summary>
        /// <param name="minLength">Minimum length of the password.</param>
        /// <param name="maxLength">Maximum length of the password.</param>
        /// <param name="useSymbols">If set to <c>true</c> then allow symbols in the result.</param>
        /// <param name="useCapitals">If set to <c>true</c> then allow capital letters in the result</param>
        /// <param name="minFrequency">The minimum allowed frequency of the 
        /// character sequences (from 0.0 to 1.0) A higher value uses more 
        /// common character sequences making tha password more memorable, 
        /// at the expense of security.</param>
        /// <param name="averageFrequency">The average frequency of character 
        /// sequences.</param>
        /// <param name="useTris">If set to <c>true</c> use three character 
        /// sequence frequencies, otherwise use four character frequencies.</param>
        /// <returns></returns>
        private string GeneratePassword(
            int minLength, int maxLength,
            bool useSymbols, bool useCapitals,
            double minFrequency, double averageFrequency,
            bool useTris)
        {
            // Paramters that vary according to whether Quad or Tri letter 
            // sequences are being used.
            int min;
            double minFq, avgFq;
            if (useTris)
            {
                min = 3;
                minFq = ConstrainFrequency(m_language.MaxTrigramFrequency * minFrequency);
                avgFq = ConstrainFrequency(m_language.MaxTrigramFrequency * averageFrequency);
            }
            else
            {
                min = 4;
                minFq = ConstrainFrequency(m_language.MaxQuadgramFrequency * minFrequency);
                avgFq = ConstrainFrequency(m_language.MaxQuadgramFrequency * averageFrequency);
            }

            // Sanitise lengths
            if (minLength < min)
            {
                minLength = min;
            }
            if (maxLength < minLength)
            {
                maxLength = minLength;
            }

            String password;
            String charsOnly;
            int sum;
            do
            {
                // Password must contain at least three letters or no 
                // frequency data will exist for the sequence.
                do
                {
                    password = GetCharacters(minLength, maxLength, m_language.GetCharacterSet(useSymbols, useCapitals));
                    Regex alphaNum = new Regex(@"[0-9!#$%&*+-/=_]");
                    charsOnly = alphaNum.Replace(password.ToLower(), "");
                } while (charsOnly.Length <= 3);

                sum = 0;
                char[] sequence = new char[] { '\x00', '\x00', '\x00', '\x00' };

                // Reject words that contain character sequences that occur 
                // less often than specified
                foreach (char c in charsOnly.ToCharArray())
                {
                    sequence[3] = c;
                    // filled the sequence?
                    if ((int)sequence[0] > 0)
                    {
                        int frequency = m_language.Frequency(sequence);
                        if (frequency < minFq)
                        {
                            // try again
                            break;
                        }
                        sum += frequency;
                    }
                    // rotate the chars towards the front of the sequence
                    sequence[0] = sequence[1];
                    sequence[1] = sequence[2];
                    sequence[2] = sequence[3];
                }
            } while (IsRude(charsOnly) || ((sum / charsOnly.Length) < avgFq));
            return password;
        }

        /// <summary>
        /// Filters undesirable words.
        /// </summary>
        /// <param name="word">The word.</param>
        /// <returns>
        /// 	<c>true</c> if the specified word is rude; otherwise, <c>false</c>.
        /// </returns>
        public bool IsRude(String word)
        {
            return IsRude(word, m_language.Filter);
        }

        /// <summary>
        /// Determines whether the specified word is rude. Allows the user to 
        /// specify a regular expression pattern as a filter.
        /// </summary>
        /// <param name="word">The word.</param>
        /// <param name="filter">The regular expression pattern that will be 
        /// used to filter the resulting words.</param>
        /// <returns>
        /// 	<c>true</c> if the specified word is rude; otherwise, <c>false</c>.
        /// </returns>
        public bool IsRude(String word, String filter)
        {
            Regex pattern = new Regex(filter);
            Match m = pattern.Match(word);
            return m.Success;
        }

        /// <summary>
        /// Gets the name of the language used by this password generator.
        /// </summary>
        /// <value>The name of the language in use by the generator.</value>
        public String GetLanguageName
        {
            get
            {
                return m_language.Name;
            }
        }

        /// <summary>
        /// Gets or sets the language.
        /// </summary>
        /// <value>The language.</value>
        public ILanguage Language
        {
            get
            {
                return m_language;
            }
            set
            {
                m_language = value;
            }
        }

    }
}
